"""Top-level package for GISMO."""

__author__ = """Fabien Mathieu"""
__email__ = 'fabien.mathieu@normalesup.org'
__version__ = '0.4.3'


from gismo.corpus import Corpus
from gismo.embedding import Embedding
from gismo.diteration import DIteration
from gismo.gismo import Gismo, XGismo



import heapq
import numpy as np
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.feature_extraction.text import CountVectorizer
from scipy.sparse import vstack, csr_matrix
from numba import njit

from gismo.corpus import Corpus, toy_source_text
from gismo.embedding import Embedding
from gismo.parameters import RESOLUTION, WIDE


class Cluster:
    """
    The 'Cluster' class is used for internal representation of hierarchical cluster. It stores
    the attributes that describe a clustering structure and provides cluster basic addition
    for merge operations.

    Parameters
    ----------
    indice: int
        Index of the head (main element) of the cluster.
    rank: int
        The ranking order of a cluster.
    vector: :class:`~scipy.sparse.csr_matrix`
        The vector representation of the cluster.

    Attributes
    ----------
    indice: int
        Index of the head (main element) of the cluster.
    rank: int
        The ranking order of a cluster.
    vector: :class:`~scipy.sparse.csr_matrix`
        The vector representation of the cluster.
    intersection_vector: :class:`~scipy.sparse.csr_matrix` (deprecated)
        The vector representation of the common points of a cluster.
    members: list of int
        The indices of the cluster elements.
    focus: float in range [0.0, 1.0]
        The consistency of the cluster (higher focus means that elements are more similar).
    children: list of :class:`~gismo.clustering.Cluster`
        The subclusters.

    Examples
    ---------
    >>> c1 = Cluster(indice=0, rank=1, vector=csr_matrix([1.0, 0.0, 1.0]))
    >>> c2 = Cluster(indice=5, rank=0, vector=csr_matrix([1.0, 1.0, 0.0]))
    >>> c3 = c1+c2
    >>> c3.members
    [0, 5]
    >>> c3.indice
    5
    >>> c3.vector.toarray()
    array([[2., 1., 1.]])
    >>> c3.intersection_vector.toarray()
    array([[1., 0., 0.]])
    >>> c1 == sum([c1])
    True
    """

    def __init__(self, indice=None, rank=None, vector=None):
        self.indice = indice
        self.rank = rank
        self.members = [indice] if indice is not None else []
        self.focus = 1.0
        self.vector = vector
        self.intersection_vector = vector
        self.children = []

    def __add__(self, cluster):
        result = Cluster()
        if self.rank < cluster.rank:
            result.indice = self.indice
            result.rank = self.rank
        else:
            result.indice = cluster.indice
            result.rank = cluster.rank
        result.members = self.members + cluster.members
        result.focus = min(self.focus, cluster.focus)  # Don't forget external focus
        result.vector = self.vector + cluster.vector
        result.intersection_vector = self.intersection_vector.multiply(cluster.intersection_vector)
        result.children = []  # Better update sons list outside class definition
        return result

    def __radd__(self, other):
        if other == 0:
            return self
        else:
            return self.__add__(other)


def merge_clusters(cluster_list: list, focus=1.0):
    """
    Complete merge operation. In addition to the basic merge provided by
    :class:`~gismo.clustering.Cluster`, it ensures the following:

    * Consistency of focus by integrating the extra-focus (typically given by :func:`~gismo.clustering.subspace_partition`).
    * Children (the members of the list) are sorted according to their respective rank.

    Parameters
    ----------
    cluster_list: list of :class:`~gismo.clustering.Cluster`
        The clusters to merge into one cluster.
    focus: float
        Evaluation of the focus (similarity) between clusters.

    Returns
    -------
    :class:`~gismo.clustering.Cluster`
        The cluster merging the list.
    """
    if len(cluster_list) < 2:
        return cluster_list[0]
    result = sum(cluster_list)
    result.focus = min(result.focus, focus)
    result.children = sorted(cluster_list, key=lambda c: c.rank)
    return result


def subspace_partition(subspace, resolution=RESOLUTION):
    """
    Proposes a partition of the subspace that merges together vectors with a similar direction.

    Parameters
    ----------
    subspace: :class:`~numpy.ndarray`, :class:`~scipy.sparse.csr_matrix`
        A ``k x m`` matrix seen as a list of ``k`` ``m``-dimensional vectors sorted by importance order.
    resolution: float in range [0.0, 1.0]
        How strict the merging should be. ``0.0`` will merge all items together, while ``1.0``
        will only merge mutually closest items.

    Returns
    -------
    list
        A list of subsets that form a partition. Each subset is represented by a pair ``(p, f)``. ``p``
        is the set of indices of the subset, ``f`` is the typical similarity of the partition (called `focus`).
    """
    # Applying a square distortion to resolution gives a more linear behavior in practice.
    resolution = 2 * resolution - resolution ** 2
    n, _ = subspace.shape
    similarity_matrix = cosine_similarity(subspace, subspace) - 2 * np.identity(n)
    similarity = np.max(similarity_matrix, axis=0)
    closest = np.argmax(similarity_matrix, axis=0)
    local_similarity = [similarity[closest[i]] for i in range(n)]
    partition = [{i} for i in range(n)]
    heads = np.arange(n)
    cluster_similarity = np.ones(n)
    for i in range(n):
        for j in range(n):
            if i == j:
                continue
            sij = similarity_matrix[i, j]
            master = min(heads[i], heads[j])
            slave = max(heads[i], heads[j])
            if sij >= resolution * local_similarity[i]:
                cluster_similarity[master] = min(cluster_similarity[master], sij)
                if master != slave:
                    for k in partition[slave]:
                        heads[k] = master
                    partition[master] |= partition[slave]
                    partition[slave] = set()
    return [(c, d) for c, d in zip(partition, cluster_similarity) if c]


def rec_clusterize(cluster_list: list, resolution=RESOLUTION):
    """
    Auxiliary recursive function for clustering.

    Parameters
    ----------
    cluster_list: list of :class:`~gismo.clustering.Cluster`
        Current aggregation state.
    resolution: float in range [0.0, 1.0]
        Sets the lazyness of aggregation. A 'resolution' set to 0.0 yields a one-step clustering
        (*star* structure), while a 'resolution ' set to 1.0 yields, up to tie similarities, a binary tree
        (*dendrogram*).

    Returns
    -------
    list of :class:`~gismo.clustering.Cluster`

    """
    if len(cluster_list) == 1:
        return cluster_list
    else:
        partition = subspace_partition(vstack([c.vector for c in cluster_list]), resolution)
        return rec_clusterize([merge_clusters([cluster_list[i] for i in p[0]], p[1]) for p in partition],
                              resolution)


def subspace_clusterize(subspace, resolution=RESOLUTION, indices=None):
    """
    Converts a subspace (matrix seen as a list of vectors) to a Cluster object (hierarchical clustering).

    Parameters
    ----------
    subspace: :class:`~numpy.ndarray`, :class:`~scipy.sparse.csr_matrix`
        A ``k x m`` matrix seen as a list of ``k`` ``m``-dimensional vectors sorted by importance order.
    resolution: float in range [0.0, 1.0]
        Sets the lazyness of aggregation. A 'resolution' set to 0.0 yields a one-step clustering
        (*star* structure), while a 'resolution ' set to 1.0 yields, up to tie similarities, a binary tree
        (*dendrogram*).
    indices: list, optional
        Indicates the index for each element of the subspace. Used when 'subspace'
        is extracted from a larger space (e.g. X or Y). If not set, indices are set to ``range(k)``.

    Returns
    -------
    Cluster
        A cluster whose leaves are the `k` vectors from 'subspace'.

    Example
    _________
    >>> corpus = Corpus(toy_source_text)
    >>> vectorizer = CountVectorizer(dtype=float)
    >>> embedding = Embedding(vectorizer=vectorizer)
    >>> embedding.fit_transform(corpus)
    >>> subspace = embedding.x[1:, :]
    >>> cluster = subspace_clusterize(subspace)
    >>> len(cluster.children)
    2
    >>> cluster = subspace_clusterize(subspace, resolution=.02)
    >>> len(cluster.children)
    4
    """
    if indices is None:
        n, _ = subspace.shape
        indices = range(n)
    return rec_clusterize([Cluster(indice=r, rank=i, vector=subspace[i, :]) for i, r in enumerate(indices)],
                          resolution)[0]


class Covering:
    def __init__(self):
        self.result = []
        self.heap = []
        self.used = set()

    def set(self, cluster):
        self.result = []
        self.heap = []
        self.used = set()
        self.push(cluster)

    def push(self, cluster):
        heapq.heappush(self.heap, (cluster.focus, cluster.rank, cluster))

    def update(self, item):
        if item not in self.used:
            self.used.add(item)
            self.result.append(item)

    def pop(self):
        return heapq.heappop(self.heap)[2]

    def core(self, cluster):
        self.set(cluster)
        while len(self.heap) > 0:
            c = self.pop()
            self.update(c.indice)
            for child in c.children:
                self.push(child)
        return self.result

    def wide(self, cluster):
        self.set(cluster)
        while len(self.heap) > 0:
            c = self.pop()
            for child in c.children:
                self.push(child)
                self.update(child.indice)
        return self.result


def covering_order(cluster, wide=WIDE):
    """
    Uses a hierarchical cluster to provide an ordering of the items that mixes rank and coverage.

    This is done by exploring all cluster and subclusters by increasing similarity and rank (lexicographic order).
    Two variants are proposed:

    * `Core`: for each cluster, append its representant to the list if new. Central items tend to have better rank.
    * `Wide`: for each cluster, append its children representants to the list if new. Marginal items tend to have better rank.

    Parameters
    ----------
    cluster: :class:`~gismo.clustering.Cluster`
        The cluster to explore.
    wide: :class:`bool`
        Use Wide (``True``) or Core (``False``) variant.

    Returns
    -------
    list of int
        Sorted indices of the items of the cluster.
    """
    if wide:
        return Covering().wide(cluster)
    else:
        return Covering().core(cluster)


@njit
def subspace_distortion(indices, data, relevance, distortion: float):
    """
    Apply inplace distortion of a subspace with relevance.

    Parameters
    ----------
    indices: :class:`~numpy.ndarray`
        Indice attribute of the subspace :class:`~scipy.sparse.csr_matrix`.
    data: :class:`~numpy.ndarray`
        Data attribute of the subspace :class:`~scipy.sparse.csr_matrix`.
    relevance: :class:`~numpy.ndarray`
        Relevance values in the embedding space.
    distortion: float in [0.0, 1.0]
        Power applied to relevance for distortion.
    """
    for i, indice in enumerate(indices):
        data[i] *= relevance[indice] ** distortion


def get_sim(csr, arr):
    """
    Simple similarity computation between csr_matrix and ndarray.

    Parameters
    ----------
    csr: :class:`~scipy.sparse.csr_matrix`
    arr: :class:`~numpy.ndarray`

    Returns
    -------
    float
    """
    return csr.dot(arr)[0]/np.linalg.norm(csr.data)/np.linalg.norm(arr)



#!/usr/bin/env python

"""Tests for `gismo` package."""

from pytest import fixture
import tempfile

from sklearn.feature_extraction.text import CountVectorizer

from gismo.common import toy_source_dict
from gismo.corpus import Corpus
from gismo.embedding import Embedding
from gismo.gismo import Gismo, XGismo


@fixture()
def my_gismo():
    corpus = Corpus(toy_source_dict, lambda x: x['content'])
    vectorizer = CountVectorizer(dtype=float)
    embedding = Embedding(vectorizer=vectorizer)
    embedding.fit_transform(corpus)
    gismo = Gismo(corpus, embedding)
    gismo.parameters.distortion = 0.0
    gismo.rank("Gizmo")
    return gismo


def test_default_cluster_document_post(my_gismo):
    cluster = my_gismo.get_documents_by_cluster(k=5)
    assert f"{cluster['focus']:.2f}" == "0.04"
    assert len(cluster["children"]) == 2


def test_default_cluster_features_post(my_gismo):
    cluster = my_gismo.get_features_by_cluster(k=10)
    assert cluster['feature'] == "mogwaï"
    assert len(cluster["children"]) == 2


def test_default_no_post(my_gismo):
    indices = my_gismo.get_documents_by_rank(k=3, post=False)
    assert list(indices) == [0, 3, 4]
    indices = my_gismo.get_features_by_rank(post=False)
    assert list(indices) == [18, 10, 14]
    indices = my_gismo.get_documents_by_coverage(k=3, post=False)
    assert list(indices) == [0, 3, 1]
    indices = my_gismo.get_features_by_coverage(post=False)
    assert list(indices) == [18, 12, 10]


def test_io_gismo(my_gismo):
    with tempfile.TemporaryDirectory() as tmpdirname:
        my_gismo.dump(filename="mygismo", path=tmpdirname)
        gismo2 = Gismo.load(filename="mygismo", path=tmpdirname)
    gismo2.rank("sentence")
    assert len(gismo2.get_documents_by_rank()) == 2


def test_io_xgismo(my_gismo):
    xgismo = XGismo(my_gismo.embedding, my_gismo.embedding)
    with tempfile.TemporaryDirectory() as tmpdirname:
        xgismo.dump(filename="mygismo", path=tmpdirname)
        xgismo2 = XGismo.load(filename="mygismo", path=tmpdirname)
    xgismo2.rank("gizmo")
    assert xgismo2.get_documents_by_rank() == ['mogwaï', 'gizmo', 'is']




#!/usr/bin/env python
# coding: utf-8
#
# GISMO: a Generic Information Search with a Mind of its Own

import numpy as np

from scipy.sparse import csr_matrix
from numba import njit
from sklearn.feature_extraction.text import CountVectorizer

from gismo.common import MixInIO, toy_source_text
from gismo.corpus import Corpus


# 1-norm for diffusion (input is X or Y indptr and data, inplace modification)
@njit
def l1_normalize(indptr, data):
    """
    Computes L1 norm on sparse embedding (x or y) and applies inplace normalization.

    Parameters
    ----------
    indptr: :class:`~numpy.ndarray`
        Pointers of the embedding (e.g. x.indptr).
    data: :class:`~numpy.ndarray`
        Values of the embedding  (e.g. x.data).

    Returns
    -------
    l1_norm: :class:`~numpy.ndarray`
        L1 norms of all vectors of the embedding before normalization.
    """

    """
    Normalize inplace the embedding (X or Y)
    :param indptr: the pointers of the vector
    :param data:  the data of the vector
    :return: the l1 norm. Data is normalized in place
    """
    n = len(indptr) - 1
    l1_norms = np.zeros(n)
    for i in range(n):
        l1_norms[i] = np.sum(data[indptr[i]:indptr[i + 1]])
        if l1_norms[i] > 0:
            data[indptr[i]:indptr[i + 1]] /= l1_norms[i]
    return l1_norms


# Note: the use of external embedding breaks a symmetry between X and Y. IDF needs to be stored if one wants to switch.

# ITF transformation
@njit
def itf_fit_transform(indptr, data, m):
    """
    Applies inplace Inverse-Term-Frequency transformation on sparse embedding x.

    Parameters
    ----------
    indptr: :class:`~numpy.ndarray`
        Pointers of the embedding (e.g. x.indptr).
    data: :class:`~numpy.ndarray`
        Values of the embedding  (e.g. x.data).
    m: int
        Number of features
    """
    n = len(indptr) - 1
    log_m = np.log(1 + m)
    for i in range(n):
        data[indptr[i]:indptr[i + 1]] *= log_m - np.log(1 + indptr[i + 1] - indptr[i])


# IDF computation
@njit
def idf_fit(indptr, n):
    """
    Computes the Inverse-Document-Frequency vector on sparse embedding y.

    Parameters
    ----------
    indptr: :class:`~numpy.ndarray`
        Pointers of the embedding y (e.g. y.indptr).
    n: int
        Number of documents.

    Returns
    -------
    idf_vector: :class:`~numpy.ndarray`
        IDF vector of size `m`.
    """
    m = len(indptr) - 1
    idf_vector = np.log(1 + n) * np.ones(m)
    for i in range(m):
        idf_vector[i] -= np.log(1 + indptr[i + 1] - indptr[i])
    return idf_vector


# IDF transformation
@njit
def idf_transform(indptr, data, idf_vector):
    """
    Applies inplace Inverse-Document-Frequency transformation on sparse embedding y.

    Parameters
    ----------
    indptr: :class:`~numpy.ndarray`
        Pointers of the embedding y (e.g. y.indptr).
    data: :class:`~numpy.ndarray`
        Values of the embedding y (e.g. y.data).
    idf_vector: :class:`~numpy.ndarray`
        IDF vector of the embedding, obtained from :func:`~gismo.embedding.idf_fit`.
    """
    m = len(indptr) - 1
    for i in range(m):
        data[indptr[i]:indptr[i + 1]] *= idf_vector[i]


@njit
def query_shape(indices, data, idf):
    """
    Applies inplace logarithmic smoothing, IDF weighting, and normalization
    to the output of the
    :class:`~sklearn.feature_extraction.text.CountVectorizer`
    :meth:`~sklearn.feature_extraction.text.CountVectorizer.transform` method.

    Parameters
    ----------
    indices: :class:`~numpy.ndarray`
        Indice attribute of the :class:`~scipy.sparse.csr_matrix` obtained from
        :meth:`~sklearn.feature_extraction.text.CountVectorizer.transform`.
    data: :class:`~numpy.ndarray`
        Data attribute of the :class:`~scipy.sparse.csr_matrix` obtained from
        :meth:`~sklearn.feature_extraction.text.CountVectorizer.transform`.
    idf: :class:`~numpy.ndarray`
        IDF vector of the embedding, obtained from :func:`~gismo.embedding.idf_fit`.

    Returns
    -------
    norm: float
        The norm of the vector before normalization.
    """
    # log normalization
    data[:] = 1 + np.log(data)
    # IdF
    for i, indice in enumerate(indices):
        data[i] *= idf[indice]
    # Normalization
    norm = np.sum(data)
    if norm > 0:
        data[:] /= norm
    return norm


def auto_vect(corpus=None):
    """
    Creates a default :class:`~sklearn.feature_extraction.text.CountVectorizer`
    compatible with the
    :class:`~gismo.embedding.Embedding` constructor.
    For not-too-small corpi, a slight frequency-filter is applied.

    Parameters
    ----------
    corpus: :class:`~gismo.corpus.Corpus`, optional
        The corpus for which the
        :class:`~sklearn.feature_extraction.text.CountVectorizer`
        is intended.

    Returns
    -------
    :class:`~sklearn.feature_extraction.text.CountVectorizer`
        A :class:`~sklearn.feature_extraction.text.CountVectorizer`
        object compatible with the
        :class:`~gismo.embedding.Embedding` constructor.
    """
    n = len(corpus) if corpus is not None else 1
    (min_df, max_df) = (3, .15) if n > 100 else (1, 1.0)
    return CountVectorizer(
        min_df=min_df,
        max_df=max_df,
        ngram_range=(1, 1),
        stop_words="english",
        dtype=float
    )


class Embedding(MixInIO):
    """
    This class leverages the :class:`~sklearn.feature_extraction.text.CountVectorizer`
    class to build the dual embedding of a :class:`~gismo.corpus.Corpus`.

    * Documents are embedded in the space of features;
    * Features are embedded in the space of documents.

    See the examples and methods below for all usages of the class.

    Parameters
    ----------
    vectorizer: :class:`~sklearn.feature_extraction.text.CountVectorizer`, optional
                Custom :class:`~sklearn.feature_extraction.text.CountVectorizer`
                to override default behavior (recommended).
                Having a :class:`~sklearn.feature_extraction.text.CountVectorizer`
                adapted to the :class:`~gismo.corpus.Corpus` is good practice.
    """

    def __init__(self, vectorizer=None):
        self.vectorizer = vectorizer
        self.n = 0  # Number of documents
        self.m = 0  # Number of features
        self.x = None  # TF-IDTF X embedding of documents into features, normalized
        self.x_norm = None  # memory of X norm for hierarchical merge
        self.y = None  # Y embedding of features into documents
        self.y_norm = None  # memory of Y norm for hierarchical merge
        self.idf = None  # idf vector
        self.features = None  # vocabulary list
        self._result_found = True  # keep track of projection successes
        self._query = ""  # keep track of projection query

    def fit_transform(self, corpus):
        """
        Ingest a corpus of documents.

        * If not yet set, a default :class:`~sklearn.feature_extraction.text.CountVectorizer` is created.
        * Features are computed and stored (fit).
        * Inverse-Document-Frequency weights of features are computed (fit).
        * TF-IDF embedding of documents is computed and stored (transform).
        * TF-ITF embedding of features is computed and stored (transform).

        Parameters
        ----------
        corpus: :class:`~gismo.corpus.Corpus`
            The corpus to ingest.

        Example
        -------
        >>> corpus=Corpus(toy_source_text)
        >>> embedding = Embedding()
        >>> embedding.fit_transform(corpus)
        >>> embedding.x  # doctest: +NORMALIZE_WHITESPACE
        <5x21 sparse matrix of type '<class 'numpy.float64'>'
            with 25 stored elements in Compressed Sparse Row format>
        >>> list(embedding.features[:8])
        ['blade', 'chinese', 'comparing', 'demon', 'folklore', 'gizmo', 'gremlins', 'inside']
        """
        if self.vectorizer is None:
            self.vectorizer = auto_vect(corpus)

        # THE FIT PART
        # Start with a simple CountVectorizer X
        x = self.vectorizer.fit_transform(corpus.iterate_text())
        # Release stop_words_ from vectorizer
        self.vectorizer.stop_words_ = None
        # Populate vocabulary
        self.features = self.vectorizer.get_feature_names_out()
        # Extract number of documents and features
        (self.n, self.m) = x.shape
        # PART OF TRANSFORM, MUTUALIZED: Apply sublinear smoothing
        x.data = 1 + np.log(x.data)
        # PART OF TRANSFORM, MUTUALIZED: Apply ITF transformation
        itf_fit_transform(indptr=x.indptr, data=x.data, m=self.m)
        # Compute transposed CountVectorizer Y
        self.y = x.tocsc()
        # Compute IDF
        self.idf = idf_fit(self.y.indptr, self.n)

        # THE TRANSFORM PART
        idf_transform(indptr=self.y.indptr, data=self.y.data, idf_vector=self.idf)
        # back to x
        self.x = self.y.tocsr(copy=True)
        # Transpose y
        self.y = self.y.T
        # Normalize
        self.x_norm = l1_normalize(indptr=self.x.indptr, data=self.x.data)
        self.y_norm = l1_normalize(indptr=self.y.indptr, data=self.y.data)

    def fit(self, corpus):
        """
        Learn features from a corpus of documents.

        * If not yet set, a default :class:`~sklearn.feature_extraction.text.CountVectorizer` is created.
        * Features are computed and stored.
        * Inverse-Document-Frequency weights of features are computed.

        Parameters
        ----------
        corpus: :class:`~gismo.corpus.Corpus`
            The corpus to ingest.

        Example
        -------
        >>> corpus=Corpus(toy_source_text)
        >>> embedding = Embedding()
        >>> embedding.fit(corpus)
        >>> len(embedding.idf)
        21
        >>> list(embedding.features[:8])
        ['blade', 'chinese', 'comparing', 'demon', 'folklore', 'gizmo', 'gremlins', 'inside']
        """
        assert corpus
        if self.vectorizer is None:
            self.vectorizer = auto_vect(corpus)

        # THE FIT PART
        # Start with a simple CountVectorizer X
        x = self.vectorizer.fit_transform(corpus.iterate_text())
        # Release stop_words_ from vectorizer
        self.vectorizer.stop_words_ = None
        # Populate vocabulary
        self.features = self.vectorizer.get_feature_names_out()
        # Extract number of documents (required for idf) and features (required in fit)
        (self.n, self.m) = x.shape
        # Compute transposed CountVectorizer Y
        self.y = x.tocsc()
        # Compute IDF
        self.idf = idf_fit(self.y.indptr, self.n)

    def fit_ext(self, embedding):
        """
        Use learned features from another :class:`~gismo.embedding.Embedding`.
        This is useful for the fast creation of local embeddings
        (e.g. at sentence level) out of a global embedding.

        Parameters
        ----------
        embedding: :class:`~gismo.embedding.Embedding`
                  External embedding to copy.

        Examples
        --------
        >>> corpus=Corpus(toy_source_text)
        >>> other_embedding = Embedding()
        >>> other_embedding.fit(corpus)
        >>> embedding = Embedding()
        >>> embedding.fit_ext(other_embedding)
        >>> len(embedding.idf)
        21
        >>> list(embedding.features[:8])
        ['blade', 'chinese', 'comparing', 'demon', 'folklore', 'gizmo', 'gremlins', 'inside']
        """
        self.m = embedding.m
        self.vectorizer = embedding.vectorizer
        self.idf = embedding.idf
        self.features = embedding.features

    def transform(self, corpus):
        """
        Ingest a corpus of documents using existing features.
        Requires that the embedding has been fitted beforehand.

        * TF-IDF embedding of documents is computed and stored.
        * TF-ITF embedding of features is computed and stored.

        Parameters
        ----------
        corpus: :class:`~gismo.corpus.Corpus`
            The corpus to ingest.

        Example
        -------
        >>> corpus=Corpus(toy_source_text)
        >>> embedding = Embedding()
        >>> embedding.fit_transform(corpus)
        >>> [embedding.features[i] for i in embedding.x.indices[:8]]
        ['gizmo', 'mogwaï', 'blade', 'sentence', 'sentence', 'shadoks', 'comparing', 'gizmo']
        >>> small_corpus = Corpus(["I only talk about Yoda", "Gizmo forever!"])
        >>> embedding.transform(small_corpus)
        >>> [embedding.features[i] for i in embedding.x.indices]
        ['yoda', 'gizmo']
        """
        # The fit part
        assert corpus

        # THE FIT PART
        # Start with a simple CountVectorizer X
        x = self.vectorizer.transform(corpus.iterate_text())
        # Release stop_words_ from vectorizer
        self.vectorizer.stop_words_ = None
        # Extract number of documents and features
        (self.n, _) = x.shape
        # PART OF TRANSFORM, MUTUALIZED: Apply sublinear smoothing
        x.data = 1 + np.log(x.data)
        # PART OF TRANSFORM, MUTUALIZED: Apply ITF transformation
        itf_fit_transform(indptr=x.indptr, data=x.data, m=self.m)
        # Compute transposed CountVectorizer Y
        self.y = x.tocsc()

        # THE TRANSFORM PART
        idf_transform(indptr=self.y.indptr, data=self.y.data, idf_vector=self.idf)
        # back to x
        self.x = self.y.tocsr(copy=True)
        # Transpose y
        self.y = self.y.T
        # Normalize
        self.x_norm = l1_normalize(indptr=self.x.indptr, data=self.x.data)
        self.y_norm = l1_normalize(indptr=self.y.indptr, data=self.y.data)

    def query_projection(self, query):
        """
        Project a query in the feature space.

        Parameters
        ----------
        query: :class:`str`
               Text to project.

        Returns
        --------
        z: :class:`~scipy.sparse.csr_matrix`
            result of the query projection (IDF distribution if query does not match any feature).
        success: :class:`bool`
            projection success (``True`` if at least one feature been found).

        Example
        -------
        >>> corpus=Corpus(toy_source_text)
        >>> embedding = Embedding()
        >>> embedding.fit_transform(corpus)
        >>> z, success = embedding.query_projection("Gizmo is not Yoda but he rocks!")
        >>> for i in range(len(z.data)):
        ...    print(f"{embedding.features[z.indices[i]]}: {z.data[i]}") # doctest: +ELLIPSIS
        gizmo: 0.3868528072...
        yoda: 0.6131471927...
        >>> success
        True
        >>> z, success = embedding.query_projection("That content does not intersect toy corpus")
        >>> success
        False
        """
        self._query = query
        z = self.vectorizer.transform([query])
        norm = query_shape(indices=z.indices, data=z.data, idf=self.idf)
        if norm == 0:
            z = csr_matrix(self.idf) / np.sum(self.idf)
            self._result_found = False
        else:
            self._result_found = True
        return z, self._result_found



import tempfile

from gismo.common import toy_source_dict, toy_source_text
from gismo.corpus import Corpus, CorpusList


def test_corpus_merge_border_cases():
    corpus = Corpus()
    assert corpus.source is None
    assert type(corpus.merge_new_source(new_source=['a', 'b'])) == Corpus
    assert corpus.source is None
    corpus.merge_new_source(new_source=['a', 'b'], doc2key=lambda x: x)
    assert corpus.source == ['a', 'b']


def test_corpuslist_io():
    assert type(CorpusList()) == CorpusList
    multi_corp = CorpusList([Corpus(toy_source_text, lambda x: x[:15]+"..."),
                             Corpus(toy_source_dict, lambda e: e['title'])])
    with tempfile.TemporaryDirectory() as tmp:
        multi_corp.dump(filename="test", path=tmp)
        new_corp = CorpusList.load(filename="test", path=tmp)
        assert len(new_corp) == 10
        assert [e for e in new_corp.iterate()][0] == 'Gizmo is a Mogwaï.'



import tempfile

from gismo.common import toy_source_text
from gismo.corpus import Corpus
from gismo.embedding import Embedding


def test_embedding_io():
    corpus=Corpus(toy_source_text)
    embedding = Embedding()
    embedding.fit_transform(corpus)
    assert embedding.features[3] == 'demon'
    with tempfile.TemporaryDirectory() as tmp:
        embedding.dump(filename="test", path=tmp)
        new_embedding = Embedding.load(filename="test", path=tmp)
    assert new_embedding.features[3] == 'demon'



#!/usr/bin/env python
# coding: utf-8
#
# GISMO: a Generic Information Search with a Mind of its Own

import gzip
import errno
import os
import dill as pickle
import numpy as np

from pathlib import Path


class MixInIO:
    """
    Provide basic save/load capacities to other classes.
    """

    def dump(self, filename: str, path='.', overwrite=False, compress=True):
        """
        Save instance to file.

        Parameters
        ----------
        filename: str
            The stem of the filename.
        path: :py:class:`str` or :py:class:`~pathlib.Path`, optional
            The location path.
        overwrite: bool
            Should existing file be erased if it exists?
        compress: bool
            Should gzip compression be used?

        Examples
        ----------

        >>> import tempfile
        >>> v1 = ToyClass(42)
        >>> v2 = ToyClass()
        >>> v2.value
        0
        >>> with tempfile.TemporaryDirectory() as tmpdirname:
        ...     v1.dump(filename='myfile', compress=True, path=tmpdirname)
        ...     dir_content = [file.name for file in Path(tmpdirname).glob('*')]
        ...     v2 = ToyClass.load(filename='myfile', path=Path(tmpdirname))
        ...     v1.dump(filename='myfile', compress=True, path=tmpdirname) # doctest.ELLIPSIS
        File ...myfile.pkl.gz already exists! Use overwrite option to overwrite.
        >>> dir_content
        ['myfile.pkl.gz']
        >>> v2.value
        42

        >>> with tempfile.TemporaryDirectory() as tmpdirname:
        ...     v1.dump(filename='myfile', compress=False, path=tmpdirname)
        ...     v1.dump(filename='myfile', compress=False, path=tmpdirname) # doctest.ELLIPSIS
        File ...myfile.pkl already exists! Use overwrite option to overwrite.

        >>> v1.value = 51
        >>> with tempfile.TemporaryDirectory() as tmpdirname:
        ...     v1.dump(filename='myfile', path=tmpdirname, compress=False)
        ...     v1.dump(filename='myfile', path=tmpdirname, overwrite=True, compress=False)
        ...     v2 = ToyClass.load(filename='myfile', path=tmpdirname)
        ...     dir_content = [file.name for file in Path(tmpdirname).glob('*')]
        >>> dir_content
        ['myfile.pkl']
        >>> v2.value
        51

        >>> with tempfile.TemporaryDirectory() as tmpdirname:
        ...    v2 = ToyClass.load(filename='thisfilenamedoesnotexist')
        Traceback (most recent call last):
         ...
        FileNotFoundError: [Errno 2] No such file or directory: ...
        """
        path = Path(path)
        destination = path / Path(filename).stem
        if compress:
            destination = destination.with_suffix(".pkl.gz")
            if destination.exists() and not overwrite:
                print(f"File {destination} already exists! Use overwrite option to overwrite.")
            else:
                with gzip.open(destination, "wb") as f:
                    pickle.dump(self, f)
        else:
            destination = destination.with_suffix(".pkl")
            if destination.exists() and not overwrite:
                print(f"File {destination} already exists! Use overwrite option to overwrite.")
            else:
                with open(destination, "wb") as f:
                    pickle.dump(self, f)

    @classmethod
    def load(cls, filename: str, path='.'):
        """
        Load instance from file.

        Parameters
        ----------
        filename: str
            The stem of the filename.
        path: :py:class:`str` or :py:class:`~pathlib.Path`, optional
            The location path.
        """
        path = Path(path)
        dest = path / Path(filename).with_suffix(".pkl")
        if dest.exists():
            with open(dest, 'rb') as f:
                return pickle.load(f)
        else:
            dest = dest.with_suffix('.pkl.gz')
            if dest.exists():
                with gzip.open(dest) as f:
                    return pickle.load(f)
            else:
                raise FileNotFoundError(
                    errno.ENOENT, os.strerror(errno.ENOENT), dest)


class ToyClass(MixInIO):
    def __init__(self, value=0):
        self.value = value


def auto_k(data, order=None, max_k=100, target=1.0):
    """
    Proposes a threshold k of significant values according to a relevance vector.

    Parameters
    ----------
    data: :class:`~numpy.ndarray`
        Vector with positive relevance values.
    order: list of int, optional
        Ordered indices of ``data``
    max_k: int
        Maximal number of entries to return; also number of entries used to determine threshold.
    target: float
        Threshold modulation. Higher target means less result.
        A target set to 1.0 corresponds to using the average of the max_k top values as threshold.

    Returns
    -------
    k: int
        Recommended number of values.

    Example
    --------
    >>> data = np.array([30, 1, 2, .3, 4, 50, 80])
    >>> auto_k(data)
    3
    """
    if order is None:
        order = np.argsort(-data)
    ordered_data = data[order[:max_k]]
    max_k = min(max_k, len(data))
    threshold = np.sum(ordered_data)*target/max_k
    k = int(np.sum(ordered_data >= threshold))
    return max(1, k)


toy_source_text = ['Gizmo is a Mogwaï.',
                   'This is a sentence about Blade.',
                   'This is another sentence about Shadoks.',
                   'This very long sentence, with a lot of stuff about Star Wars inside, makes at some point a side '
                   'reference to the Gremlins movie by comparing Gizmo and Yoda.',
                   'In chinese folklore, a Mogwaï is a demon.']
"""A minimal source example where items are :py:obj:`str`."""

toy_source_dict = [{'title': 'First Document', 'content': 'Gizmo is a Mogwaï.'},
                   {'title': 'Second Document', 'content': 'This is a sentence about Blade.'},
                   {'title': 'Third Document', 'content': 'This is another sentence about Shadoks.'},
                   {'title': 'Fourth Document',
                    'content': 'This very long sentence, with a lot of stuff about Star Wars inside, '
                               'makes at some point a side reference to the Gremlins movie by '
                               'comparing Gizmo and Yoda.'},
                   {'title': 'Fifth Document', 'content': 'In chinese folklore, a Mogwaï is a demon.'}]
"""A minimal source example where items are :py:obj:`dict` with keys `title` and `content`."""



import tempfile
import io
from pathlib import Path
from pytest import fixture

from gismo.datasets.dblp import Dblp

URL = "https://github.com/balouf/gismo/raw/master/gismo/datasets/acm.json.gz"


def test_dblp_download_gz():
    """
    Download a gz file (not a dblp file for this test).
    """
    with tempfile.TemporaryDirectory() as tmp:
        dblp = Dblp(dblp_url=URL, path=Path(tmp))
        dblp.download()
        dest = Path(tmp) / Path("dblp.xml.gz")
        assert dest.exists() == True



import requests
from zipfile import ZipFile
import io

# URL = "https://archive.ics.uci.edu/ml/machine-learning-databases/00217/C50.zip"
URL = "https://github.com/balouf/datasets/raw/main/C50.zip"

def get_reuters_entry(name, z):
    """
    Read the Reuters news referenced by `name` in the zip archive `z` and returns it as a dict.

    Parameters
    ----------
    name: str
        Location of the file inside the Reuters archive
    z: ZipFile
        Zipfile descriptor of the Reuters archive

    Returns
    -------
    entry: dict
        dict with keys `set` (`C50test` or `c50train`), `author`, `id`, and `content`
    """
    with z.open(name) as f:
        description = name.split("/")
        return {'set': description[0],
                'author': description[1],
                'id': description[2][:-4],
                'content': f.read().decode()}


def get_reuters_news(url=URL):
    """
    Returns a list of news from the Reuters C50 news datasets

    Acknowledgments

    Dua, D. and Graff, C. (2019). UCI Machine Learning Repository [http://archive.ics.uci.edu/ml].
    Irvine, CA: University of California, School of Information and Computer Science.

    ZhiLiu, e-mail: liuzhi8673 '@' gmail.com,
    institution: National Engineering Research Center for E-Learning, Hubei Wuhan, China

    Parameters
    ----------
    url: str
        Location of the C50 dataset

    Returns
    -------
    list
        The C50 news as a list of dict

    Example
    ---------
    Cf :py:class:`~gismo.sentencizer.Sentencizer`
    """
    r=requests.get(url)
    with ZipFile(io.BytesIO(r.content)) as z:
        return [get_reuters_entry(name, z) for name in z.namelist() if name.endswith('.txt')]



import logging

logging.basicConfig()
log = logging.getLogger("Gismo")

ALPHA = .5
"""
Default value for damping factor.
Controls the trade-off between closeness and centrality.
"""

N_ITER = 4
"""
Default value for the number of round-trip diffusions to perform.
Higher value means better precision but longer execution time.
"""

OFFSET = 1.0
"""
Default offset value.
Controls how much of the initial fluid should be deduced from the relevance.
"""

MEMORY = 0.0
"""
Default memory value.
Controls how much of previous computation is kept
when performing a new diffusion.
"""

STRETCH = 2.0
"""
Default stretch value.
When performing covering, defines the ratio between considered pages and
selected covering pages.
"""

RESOLUTION = 0.7
"""
Default resolution value.
Defines how strict the merging of cluster is during recursive clustering.
"""

MAX_K = 100
"""
Default top population size for estimating k.
"""

TARGET_K = 1.0
"""
Default threshold for estimating k.
"""

WIDE = True
"""
Default Covering behavior for covering.
True for wide variant, false for core variant.
"""

POST = True
"""
Default post policy.
If True, post function is applied on items and clusters.
"""

DISTORTION = 1.0
"""
Default distortion.
Controls how much of diteration relevance is mixed into the embedding
for similarity computation.
"""

DEFAULT_PARAMETERS = {'alpha': ALPHA,
                      'n_iter': N_ITER,
                      'offset': OFFSET,
                      'memory': MEMORY,
                      'stretch': STRETCH,
                      'resolution': RESOLUTION,
                      'max_k': MAX_K,
                      'target_k': TARGET_K,
                      'wide': WIDE,
                      'post': POST,
                      'distortion': DISTORTION
                      }
"""
Dictionary of default `runtime` :class:`~gismo.gismo.Gismo` parameters.
"""

X_DENSITY = 1000
"""Default number of documents representing a :class:`~gismo.landmarks.Landmarks` entry."""

Y_DENSITY = 1000
"""Default number of features representing a :class:`~gismo.landmarks.Landmarks` entry."""

BALANCE = 0.5
"""Default documents/features trade-off in :class:`~gismo.landmarks.Landmarks`."""

DEFAULT_LANDMARKS_PARAMETERS = {
    'stretch': STRETCH,
    'resolution': RESOLUTION,
    'max_k': MAX_K,
    'target_k': TARGET_K,
    'wide': WIDE,
    'post': POST,
    'distortion': DISTORTION,
    'x_density': X_DENSITY,
    'y_density': Y_DENSITY,
    'balance': BALANCE,
    'rank': lambda g, q: g.rank(q)
}
"""
Dictionary of default `runtime` :class:`~gismo.landmarks.Landmarks` parameters.
"""


class Parameters:
    """
    Manages :class:`~gismo.gismo.Gismo` runtime parameters. When called, an instance will yield
    a dictionary of parameters. Is also used for other Gismo classes like :class:`~gismo.landmarks.Landmarks`.

    Parameters
    ----------
    parameter_list: dict, optional
        Indicates which paramaters to manage. Default to Gismo runtime parameter.
    kwargs: dict
        Parameters that need to be distinct from default values.

    Examples
    --------

    Use default parameters.

    >>> p = Parameters()
    >>> p() # doctest: +NORMALIZE_WHITESPACE
    {'alpha': 0.5, 'n_iter': 4, 'offset': 1.0, 'memory': 0.0,
    'stretch': 2.0, 'resolution': 0.7, 'max_k': 100, 'target_k': 1.0,
    'wide': True, 'post': True, 'distortion': 1.0}

    Use default parameters with changed `stretch`.

    >>> p = Parameters(stretch=1.7)
    >>> p()['stretch']
    1.7

    Note that parameters that do not exist will be ignored and (a warning will
    be issued)

    >>> p = Parameters(strech=1.7)
    >>> p() # doctest: +NORMALIZE_WHITESPACE
    {'alpha': 0.5, 'n_iter': 4, 'offset': 1.0, 'memory': 0.0,
    'stretch': 2.0, 'resolution': 0.7, 'max_k': 100, 'target_k': 1.0,
    'wide': True, 'post': True, 'distortion': 1.0}

    You can change the value of an attribute to alter the returned parameter.

    >>> p.alpha = 0.85
    >>> p()['alpha']
    0.85

    You can also apply on-the-fly parameters by passing them when
    calling the instance.

    >>> p(resolution=0.9)['resolution']
    0.9

    Like for construction, parameters that do not exist are ignored and a warning
    is issued.

    >>> p(resolutio = .9) # doctest: +NORMALIZE_WHITESPACE
    {'alpha': 0.85, 'n_iter': 4, 'offset': 1.0, 'memory': 0.0,
    'stretch': 2.0, 'resolution': 0.7, 'max_k': 100, 'target_k': 1.0,
    'wide': True, 'post': True, 'distortion': 1.0}

    Note the possibility to store a custom set of parameters if one uses ``parameter_list`` in construction.

    >>> p = Parameters(parameter_list={'a': 1.0, 'b': True}, a=1.5)
    >>> p()
    {'a': 1.5, 'b': True}
    """

    def __init__(self, parameter_list=None, **kwargs):
        if parameter_list is None:
            parameter_list = DEFAULT_PARAMETERS
        for key, value in parameter_list.items():
            setattr(self, key, value)
        for key, value in kwargs.items():
            if hasattr(self, key):
                if value is not None:
                    setattr(self, key, value)
            else:
                log.warning(f"No parameter named {key}!")

    def __call__(self, **kwargs):
        d = {key: value for key, value in self.__dict__.items()}
        for key, value in kwargs.items():
            if key in d:
                if value is not None:
                    d[key] = value
            else:
                log.warning(f"No parameter named {key}!")
        return d
